/*
 * Part of the solution for Assignment 3 (CSC 218, Winter 2003), by
 * Stefan Bruda.
 */
#include "stree.h"
#include <iostream>
#include <string.h>
#include <libgen.h> // for basename()
using namespace std;

/***
 *** Variable for command parsing:
 ***/

/* 
 * Let's define constants for each known command:
 */
const char* STR_MAKE =    "make";
const char* STR_MEMBER =  "member";
const char* STR_DELETE =  "delete";
const char* STR_INSERT =  "insert";
const char* STR_PRINT =   "print";
const char* STR_LIST =    "list";
const char* STR_CHECK =   "check";
const char* STR_RESTORE = "restore";
const char* STR_QUIT =    "quit";

/*
 * Let's also establish numerical values for those commands (an array
 * of commands will come in handy):
 */
const int MAKE =    0;
const int MEMBER =  1;
const int DELETE =  2;
const int INSERT =  3;
const int PRINT =   4;
const int LIST =    5;
const int CHECK =   6;
const int RESTORE = 7;
const int QUIT =    8;
// error conditions:
const int UNKNOWN =   -1;  // command unknown
const int AMBIGUOUS = -2;  // ambiguous prefix
const int NO_PARAM =  -3;  // no parameter

/*
 * The NULL-terminated array of commands (indices correspond to the
 * respective numerical values of the commands).
 */
const char* commands[] = {STR_MAKE, STR_MEMBER, STR_DELETE, STR_INSERT, 
                          STR_PRINT, STR_LIST, STR_CHECK, STR_RESTORE, 
                          STR_QUIT, 0};

/*
 * The number of required arguments for each command (0 or 1).
 */
const int command_args[] = {0,1,1,1,0,0,0,0,0};


/***
 *** Now to our actual program:
 ***/

/*
 * parse_command(i,c) receives a command line (a command, followed by
 * a number of consecutive blanks, followed by a possibly empty
 * argument), returns the numerical value corresponding to c, and sets
 * i to the first character of the argument (so uppon return the
 * argument is accessible at &c[i]).  If there is no argument, i is
 * set such that &c[i] is the empty string.  As a side effect an
 * appropriate null byte is placed inside c, such that upon return c
 * will refer to the (null-terminated) command string rather than the
 * whole command line.  This is not really useful, except for internal
 * processing and for issuing error messages when appropriate in the
 * main function.
 */
int parse_command (int& arg_index, char* command) {
    size_t i = 0;
    int done = 0;
  
    // we start from the premise that there is no argument.
    arg_index = strlen(command);

    // tokenize command:
    while ( i <= strlen(command) && !done) {
        if (command[i] == ' ') {  // break original string at first blank
            command[i] = '\0';
            i++;
            while (command[i] == ' ')  // ignore any consecutive blanks...
                i++;
            arg_index = i;             // ... and set arg_index to the first
            done = 1;                  // non-blank character.
        }
        else
            i++;
    }
    // now command contains only the command itself.

    i = 0;
    int com = UNKNOWN;  // the returned command code

    // identify command:
    while (commands[i]) {
        if ( strncmp(command, commands[i], strlen(command)) == 0 ) { // got it!
            if (com != -1) // found it before?
                return AMBIGUOUS;
            else
                com = i;     // do not interrupt the cycle though (in case we
                             // have an ambiguous prefix).
        }
        i++;
    }
    if (com == UNKNOWN)
        return UNKNOWN;
    if (command[arg_index] == '\0' && command_args[com] > 0)
        return NO_PARAM;
    return com;
}

/*
 * Maximum number of backups:
 */
const int NO_BAK = 5;


/*
 * Keeps a search tree in memory.  Repeatedly reads commands from the
 * standard input and apply them to the search tree.  Terminates upon
 * receipt of command "quit" (or an unambiguous prefix thereof).
 */
int main (int argc, char** argv) {
    int arg_index;          // for parse_command (return value)
    int done = 0;           // when 1 interrupts main cycle (set by command quit)
    char command[1024];     // current command line

    // The working search tree:
    stree* working_tree = 0;

    // Declaration and initialization for the backup stack:
    stree* backups[NO_BAK];
    for (int i = 0; i < NO_BAK; i++ ) {
        backups[i] = 0;
    }

    while (!done) {
        cout << basename(argv[0]) << "% ";  // just a prompt...
        cout.flush();
        cin.getline(command,1024);
        int c = parse_command(arg_index,command);
        
        char* arg = &command[arg_index];    // just a convenient shorthand for the
                                            // argument of the current command...
        // We deal first with commands that do not necessarily require a
        // working tree:
        switch(c) { 
            // Error conditions:
        case UNKNOWN:
            cerr << "*** Command " << command << " is unknown here.\n";
            continue;
        case AMBIGUOUS:
            cerr << "*** Command " << command << " is ambiguous.\n";
            continue;
        case NO_PARAM:
            cerr << "*** Command " << command << " requires one argument, you supplied none.\n";
            cerr << "*** So I will just assume you want to see the backup stack. :-)\n";
            for (int i = 0; i < NO_BAK; i++) {
                cout << "*** Backup " << i << ":\n";
                if (backups[i])
                    cout << *(backups[i]);
                else
                    cout << "EMPTY\n";
            }
            continue;
            // Create working tree:
        case MAKE:
            delete working_tree;
            working_tree = new stree;
            continue;
            // The names say it all:
        case QUIT:
            done = 1;
            continue;
        case RESTORE:  // CHECK needs a tree, RESTORE does not...
            // make room:
            delete working_tree;
            // restore:
            if (backups[0]) { // we have what to restore?
                working_tree = backups[0];           // get the most recent backup
                for (int i = 0; i < NO_BAK - 2; i++) // shift remaining backups
                    backups[i] = backups[i+1];
                backups[NO_BAK-1] = 0;    // this one gets freed.
            }
            else // nothing to restore, create a new working tree.
                working_tree = new stree;
            continue;
        }
        // The remaining commands do not work if a working tree is not in
        // place, so we check first for such:
        if (working_tree == 0) {
            cerr << "*** " << command << ": no working tree (use `make' to create one).\n";
        }
        else  // Now we have a tree, let's do something with it:
            switch(c) {
                // MAKE, QUIT, RESTORE already taken care of.
                // The remaining are quite trivial (except perhaps for CHECK).
            case MEMBER:
                if (working_tree -> member(arg))
                    cout << "yes\n";
                else
                    cout << "no\n";
                break;
            case DELETE:
                btree::node* to_delete;
                while ( (to_delete = working_tree -> member(arg)) ) {
                    working_tree -> del(to_delete); // deep delete!
                }
                break;
            case INSERT:
                working_tree -> insert(arg); // the copying of arg done by insert.
                break;
            case PRINT:
                cout << *working_tree;
                break;
            case LIST:
                working_tree -> print(cout,1);
            case CHECK:
                // get rid of the NO_BAK-th backup first (no more room for it).
                delete backups[NO_BAK-1];        
                // shift remaining backups:
                for (int i = NO_BAK-1; i > 0; i--) {
                    backups[i] = backups[i-1];
                }
                // now we have room in backups[0] for the current tree; we
                // have to _copy_ it, since this is a backup        
                backups[0] = new stree(*working_tree);  // COPY CONSTRUCTOR at work!
                break;
            }
    }  // end main loop
    // clean up (always good, be it just to impress the prof ;-) ):
    delete working_tree;
    for (int i = 0; i < NO_BAK; i++) {
        delete backups[i];
    }
}
